package com.bergerkiller.bukkit.nolagg;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.logging.Level;

import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.World;
import org.bukkit.command.CommandSender;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Item;
import org.bukkit.entity.LivingEntity;
import org.bukkit.entity.Player;
import org.bukkit.entity.TNTPrimed;

import com.bergerkiller.bukkit.nolagg.meta.WorldMetaData;

public class PerformanceMonitor implements Runnable {
	
	public static int monitorInterval = 40;
	
	private long prevtime;
	private long prevusedmem;
	private long minmem;
	
	private static final double oldper = 0.3;
	public static final ProcessTime entityListGen = new ProcessTime(oldper, "Entity lstng", ChatColor.GREEN);
	public static final ProcessTime entitySpawnLimit = new ProcessTime(oldper, "Spwn limtr", ChatColor.DARK_GREEN);
	public static final ProcessTime itemStack = new ProcessTime(oldper, "Itm stckr", ChatColor.BLUE);
	public static final ProcessTime chunkUpdate = new ProcessTime(oldper, "Chnk updtr", ChatColor.DARK_RED);
	public static final ProcessTime chunkSend = new ProcessTime(oldper, "Chnk sending", ChatColor.RED);
	private static final ProcessTime[] times = new ProcessTime[] {entityListGen, entitySpawnLimit, itemStack, chunkUpdate, chunkSend};

	
	public static ArrayList<String> recipients = new ArrayList<String>();
	public static boolean sendConsole = false;
	public static boolean sendLog = false;
	private static final Runtime runtime = Runtime.getRuntime();
	private static int taskID = -1;
	private static BufferedWriter logger;
	private static boolean wroteHeader = false;
	public static boolean broadcastMemoryHigh = true;
	
	private static File logfile;
	
	public static void init() {
		PerformanceMonitor pm = new PerformanceMonitor();
		pm.prevtime = System.currentTimeMillis();
		pm.prevusedmem = runtime.maxMemory() - runtime.freeMemory();
		pm.minmem = pm.prevusedmem;
		taskID = NoLagg.plugin.getServer().getScheduler().scheduleSyncRepeatingTask(NoLagg.plugin, pm, monitorInterval, monitorInterval);
		//set up logger
		logfile = NoLagg.plugin.getDataFolder();
		logfile.mkdirs();
		logfile = new File(logfile + File.separator + "log.txt");
		try {
			logger = new BufferedWriter(new FileWriter(logfile, true));
			log("NoLagg enabled: " + getStamp());
			logger.flush();
		} catch (IOException ex) {
			NoLagg.log(Level.SEVERE, "Failed to initialize performance logger file stream:");
			ex.printStackTrace();
		}
	}
	public static void deinit() {
		NoLagg.plugin.getServer().getScheduler().cancelTask(taskID);
		//set up logger
		if (logger != null) {
			try {
				log("NoLagg disabled: " + getStamp());
				logger.flush();
				logger.close();
			} catch (IOException ex) {}
			logger = null;
		}
		recipients.clear();
	}
	
	private String getProgress(int length, ChatColor color) {
		if (length <= 0) return "";
		StringBuilder sb = new StringBuilder(length + 1);
		sb.append(color);
		for (int i = 0; i < length; i++) {
			sb.append('|');
		}
		return sb.toString();
	}
	private String getMemoryProgress(int length, long current, long min, long max) {
		double factor = (double) length / max;
		StringBuilder sb = new StringBuilder(length + 3);
		int used = (int) (factor * min);
		int unused = (int) (factor * (max - current));
		sb.append(getProgress(used, ChatColor.GREEN));
		sb.append(getProgress(length - unused - used, ChatColor.YELLOW));
		sb.append(getProgress(unused, ChatColor.RED));
		return sb.toString();
	}
	
    private static double round(double Rval, int Rpl) {
    	double p = Math.pow(10, Rpl);
    	return Math.round(Rval * p) / p;
    }
    
    private static int mem(double value) {
    	return mem((long) value);
    }
    private static int mem(long value) {
    	return (int) (value / 1048576);
    }
    
    public static boolean clearLog() {
    	if (logger != null) {
    		try {
    			logger.close();
    		} catch (Exception ex) {
    			return false;
    		}
    	}
    	if (logfile.delete()) {
    		try {
    			logger = new BufferedWriter(new FileWriter(logfile, true));
    			wroteHeader = false;
    			return true;
    		} catch (Exception ex) {
    			return false;
    		}
    	} else {
    		return false;
    	}
    }
    
    private static void log(String message) throws IOException {
		logger.write(message);
		logger.newLine();
    }
    public static boolean writeLog(String message) {
    	try {
    		log(message);
    		return true;
    	} catch (IOException ex) {
    		return false;
    	}
    }
    
    private static String getTime() {
    	final SimpleDateFormat sdf = new SimpleDateFormat("H:mm:ss");
    	return sdf.format(Calendar.getInstance().getTime());
    }
    private static String getStamp() {
    	final SimpleDateFormat sdf = new SimpleDateFormat("yyyy.MM.dd H:mm:ss");
    	return sdf.format(Calendar.getInstance().getTime());
    }
    
    private static String getColumn(String text) {
    	return " " + text + "		|";
    }
    
    public static class ProcessTime {
    	private final double useoldper;
    	public ProcessTime(final double useoldper, String name, ChatColor color) {
    		this.useoldper = useoldper;
    		this.name = name;
    		this.color = color;
    	}
    	public void update() {
    		this.update(1);
    	}
    	public void update(int tickspan) {
    		this.update(StopWatch.next(), tickspan);
    	}
    	public void update(final double elapsedtime, int tickspan) {
    		this.duration = Util.useOld(this.duration, elapsedtime / tickspan, this.useoldper);
    	}
    	public double duration;
    	public String name;
    	public ChatColor color;
    	public int barlength = 0;
    	public String toString() {
    		return this.color + "[" + this.name + "]";
    	}
    }

	@Override
	public void run() {	
		long time = System.currentTimeMillis();
		double elapsedtimesec = (double) (time - prevtime) / 1000;
		long totalmem = runtime.totalMemory();
		long usedmem = totalmem - runtime.freeMemory();
		long diff = usedmem - prevusedmem;
		if (diff < 0) {
			minmem = usedmem;
		}
		if (sendLog || sendConsole || recipients.size() > 0) {
			String cpacketbuff = BufferedPacket51MapChunk.getBufferSizeName();
			int mobcount = 0;
			int entitycount = 0;
			int itemcount = 0;
			int tntcount = 0;
			int playercount = 0;
			for (World w : Bukkit.getServer().getWorlds()) {
				for (Entity e : w.getEntities()) {
					if (e instanceof LivingEntity) {
						if (e instanceof Player) {
							playercount++;
						} else {
							mobcount++;
						}
					} else if (e instanceof Item) {
						itemcount++;
					} else if (e instanceof TNTPrimed) {
						tntcount++;
					}
					entitycount++;
				}
			}
			double tps = monitorInterval / elapsedtimesec;
			int hiddenItems = WorldMetaData.getTotalHiddenItemCount();
			int savesize = ChunkScheduler.size();
			int totalchunkcount = WorldMetaData.getTotalChunkCount();
			int totaluchunkcount = WorldMetaData.getTotalUnloadableChunkCount();
			long lchunks = NLWorldListener.loadedChunks - NLWorldListener.genChunks;
			if (sendLog && logger != null) {
				try {
					if (!wroteHeader) {
						wroteHeader = true;
						String columns = "";
						columns += "| Time		| Tick rate		| Total Memory	| Static Memory	| Dynamic Memory	";
						columns += "| Memory Write Rate	| Total Chunks	| Unloadable Chunks	";
						columns += "| Chunks Loaded	| Chunks generated	| Chunks Unloaded	| Chunks To Save	| Buffered TNT	| Buffered Items	"; 
						columns += "| Chunk Packets	| Entities		| Mobs		| Items		| TNT		| Players		|";
						log(columns);
					}
					String msg = "";
					msg += "|" + getColumn(getTime());
					msg += getColumn(String.valueOf(round(tps, 1)));
					msg += getColumn(mem(totalmem) + " MB");
					msg += getColumn(mem(minmem) + " MB");
					msg += getColumn(mem(usedmem - minmem) + " MB");
					if (diff <= 0) {
						msg += getColumn("GC");
					} else {
						msg += getColumn(mem(diff / elapsedtimesec) + " MB/s");
					}
					msg += getColumn(String.valueOf(totalchunkcount));
					msg += getColumn(String.valueOf(totaluchunkcount));
					msg += getColumn(String.valueOf(lchunks));
					msg += getColumn(String.valueOf(NLWorldListener.genChunks));
					msg += getColumn(String.valueOf(NLUnloadListener.unloadedChunks));
					msg += getColumn(String.valueOf(savesize));
					msg += getColumn(String.valueOf(TnTHandler.getBufferCount()));
					msg += getColumn(String.valueOf(hiddenItems));
					msg += getColumn(cpacketbuff);
					msg += getColumn(String.valueOf(entitycount));
					msg += getColumn(String.valueOf(mobcount));
					msg += getColumn(String.valueOf(itemcount));
					msg += getColumn(String.valueOf(tntcount));
					msg += getColumn(String.valueOf(playercount));
					log(msg);
					logger.flush();
				} catch (IOException ex) {
					NoLagg.log(Level.SEVERE, "Logging disabled:");
					ex.printStackTrace();
					try {
						logger.close();
					} catch (IOException e) {}
					logger = null;
				}
			}
			
			if (sendConsole) {
				CommandSender s = Bukkit.getServer().getConsoleSender();
				//Line
				s.sendMessage("-");
				//memory
				String mem = "Memory: " + mem(minmem) + "/" + mem(totalmem) + " MB (+" + mem(usedmem - minmem) + " modified)";
				mem += "(+" + mem(diff / elapsedtimesec) + " MB/s)";
				s.sendMessage(mem);
				//chunks
				mem = "Chunks: ";
				mem += totalchunkcount + " [" + totaluchunkcount + " Unloadable]";
				mem += " [+" + NLWorldListener.loadedChunks + "]";
				mem += " [-" + NLUnloadListener.unloadedChunks + "]";
				mem += " [" + savesize + " left to save]";
				s.sendMessage(mem);
				//Entities
				mem = "Entities: " + entitycount + " [" + mobcount + " mobs]";
				mem += " [" + itemcount + " items] [" + tntcount + " mobile TNT]";
				s.sendMessage(mem);
				//Buffering
				mem = "Buffers: [" + TnTHandler.getBufferCount() + " TNT] [" + hiddenItems + " items] [" + cpacketbuff + " Chunk Packets]";;
				s.sendMessage(mem);
				//Tick times
				mem = "Ticks per second: " + round(tps, 1) + " [" + round(tps * 5, 0) + "%]";
				s.sendMessage(mem);
			}
			if (recipients.size() > 0) {
				ArrayList<String> messages = new ArrayList<String>(6);
				//Line
				messages.add("");
				
				String mem;
				//Global update times
				double totaltime = 0;
				for (ProcessTime ptime : times) totaltime += ptime.duration;
				if (totaltime > 0) {
					mem = "";
					for (ProcessTime ptim : times) mem += ptim.toString();
					messages.add(mem);

					final int length = 50;
					//get bar lengths
					int totallength = 0;
					for (ProcessTime ptime : times) {
						totallength += (ptime.barlength = (int) (length * ptime.duration / totaltime));
					}
					int i = 0;
					while (totallength > length) {
						times[i].barlength--;
						totallength--;
						i++;
						if (i > times.length - 1) i = 0;
					}
					while (totallength < length) {
						times[i].barlength++;
						totallength++;
						i++;
						if (i > times.length - 1) i = 0;
					}
					//get message
					
					mem = ChatColor.YELLOW + "Update: ";
					for (ProcessTime ptime : times) {
						mem += getProgress(ptime.barlength, ptime.color);
					}
					mem += ChatColor.YELLOW + " (" + Util.round(totaltime, 1) + " MS/tick)";
					messages.add(mem);
				}
				
				//Memory
				mem = ChatColor.YELLOW + "Memory: ";
				mem += getMemoryProgress(50, usedmem,  minmem, totalmem);
				mem += ChatColor.YELLOW + " " + mem(minmem) + "/" + mem(totalmem) + " MB ";
				if (diff > 0) {
					mem += ChatColor.RED + "(+" + mem(diff / elapsedtimesec) + " MB/s)";
				} else {
					mem += ChatColor.GREEN + "(GC)";
				}
				messages.add(mem);
				//Chunks
				mem = ChatColor.YELLOW + "Chunks: ";
				mem += totalchunkcount + " [" + totaluchunkcount + " U]";
				mem += ChatColor.GREEN + " [+" + lchunks + "]";
				mem += ChatColor.YELLOW + " [+" + NLWorldListener.genChunks + "]";
				mem += ChatColor.RED + " [-" + NLUnloadListener.unloadedChunks + "]";
				mem += ChatColor.YELLOW + " [" + savesize + " left to save]";
				messages.add(mem);
				//Entities
				mem = ChatColor.YELLOW + "Entities: ";
				mem += entitycount + " [" + mobcount + " mobs] [" + itemcount + " items] [" + tntcount + " TNT] [" + playercount + " players]";
				messages.add(mem);
				//Buffering
				mem = ChatColor.YELLOW + "Buffers: [" + TnTHandler.getBufferCount() + " TNT] [" + hiddenItems + " items] [" + cpacketbuff + " Chunk Packets]";
				messages.add(mem);
				//Tick times
				mem = ChatColor.YELLOW + "Ticks per second: " + round(tps, 1) + " [" + round(tps * 5, 0) + "%]";
				messages.add(mem);
			
				//average send rate
				double avgrate = 0;
				if (ChunkSendQueue.enabled) {
					avgrate = ChunkSendQueue.getAverageRate();
				}
				
				int i = 0;
				while (i < recipients.size()) {
					String name = recipients.get(i);
					Player p = Bukkit.getServer().getPlayer(name);
					if (p != null) {
						for (String msg : messages) {
							p.sendMessage(msg);
						}
						//send individual sending rate
						if (ChunkSendQueue.enabled) {
							ChunkSendQueue queue = ChunkSendQueue.bind(p);
							int tosend = Util.chunkarea - queue.size();
							String msg = ChatColor.YELLOW + "Chunk sending: " + (tosend * 100 / Util.chunkarea) + "% of " + Util.chunkarea;
							msg += " at " + Util.round(queue.getRate(), 2) + " chunks/tick (" + Util.round(avgrate, 2) + " avg)";
							p.sendMessage(msg);
						}
						i++;
					} else {
						recipients.remove(i);
					}
				}
			}
		}

		NLWorldListener.loadedChunks = 0;
		NLUnloadListener.unloadedChunks = 0;
		NLWorldListener.genChunks = 0;
		prevtime = time;
		this.prevusedmem = usedmem;
	}

}
